"""Tests for FilService.should_update property."""

from datetime import UTC, datetime, timedelta
from pathlib import Path

import pytest

from avwx_swift.fil import FilService


def make_service(server_time: datetime | None = None, updated: datetime | None = None) -> FilService:
    """Helper to make a FilService with specific times."""
    service = FilService(url="example.com", user="user", cert_path=Path("/tmp/cert.pem"))
    service.server_time = server_time
    service.updated = updated
    return service


def test_should_update_no_server_time() -> None:
    """Should raise if server_time is None."""
    service = make_service(server_time=None, updated=None)
    with pytest.raises(ValueError, match="Server time has not been fetched."):
        _ = service.should_update


def test_should_update_no_updated() -> None:
    """Should return True if never updated."""
    service = make_service(server_time=datetime.now(UTC), updated=None)
    assert service.should_update is True


def test_should_update_updated_before_server_time() -> None:
    """Should return True if server is more recent than last update."""
    now = datetime.now(UTC)
    before = now - timedelta(hours=1)
    service = make_service(server_time=now, updated=before)
    assert service.should_update is True


def test_should_update_updated_equal_server_time() -> None:
    """Should return False if server time equals last update."""
    now = datetime.now(UTC)
    service = make_service(server_time=now, updated=now)
    assert service.should_update is False


def test_should_update_updated_after_server_time() -> None:
    """Should return False if last update is more recent than server time."""
    now = datetime.now(UTC)
    after = now + timedelta(hours=1)
    service = make_service(server_time=now, updated=after)
    assert service.should_update is False



"""Solace JMS-like service implementation using Solace Python API."""

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from solace.messaging.config.missing_resources_creation_configuration import MissingResourcesCreationStrategy
from solace.messaging.config.solace_properties import (
    authentication_properties,
    service_properties,
    transport_layer_properties,
)
from solace.messaging.messaging_service import MessagingService
from solace.messaging.receiver.message_receiver import MessageHandler
from solace.messaging.receiver.persistent_message_receiver import PersistentMessageReceiver
from solace.messaging.resources.queue import Queue

_DEFAULT_CERT_PATH = Path(__file__).parent / "cacerts"


log = logging.getLogger(__name__)


@dataclass
class JmsConfig:
    """Config values for data push service."""

    username: str
    password: str
    queue_name: str
    connection_factory: str
    url: str
    message_vpn: str

    thread_count: int = 4
    ssl_trust_store: Path = _DEFAULT_CERT_PATH
    jndi_connection_retries: int = 0  # -1 for infinite retries


class JmsService:
    """Solace JMS-like service implementation using Solace Python API."""

    config: JmsConfig
    service: MessagingService
    receiver: PersistentMessageReceiver | None = None

    def __init__(self, config: JmsConfig, event_handler: Any = None):
        self.config = config

        self.service = (
            MessagingService.builder()
            .from_properties(
                {
                    authentication_properties.SCHEME_BASIC_PASSWORD: config.password,
                    authentication_properties.SCHEME_BASIC_USER_NAME: config.username,
                    service_properties.VPN_NAME: config.message_vpn,
                    transport_layer_properties.CONNECTION_RETRIES: config.jndi_connection_retries,
                    transport_layer_properties.HOST: config.url,
                    transport_layer_properties.RECONNECTION_ATTEMPTS: config.jndi_connection_retries,
                    # transport_layer_security_properties.TRUST_STORE_PATH: config.ssl_trust_store.absolute().as_posix(),
                    # transport_layer_properties.RECONNECTION_ATTEMPTS_WAIT_INTERVAL: 3000,
                    # transport_layer_security_properties.CERT_VALIDATED: False,
                    # transport_layer_security_properties.CERT_REJECT_EXPIRED: False,
                    # transport_layer_security_properties.CERT_VALIDATE_SERVERNAME: False,
                }
            )
            .build()
        )

        # Blocking connect thread
        log.debug("Connecting to Messaging Service at %s...", self.config.url)
        self.service.connect()
        log.debug("Messaging Service connected? %s", self.service.is_connected)

        # Event Handling for the messaging service
        if event_handler:
            if hasattr(event_handler, "on_reconnected"):
                self.service.add_reconnection_listener(event_handler)
            if hasattr(event_handler, "on_reconnecting"):
                self.service.add_reconnection_attempt_listener(event_handler)
            if hasattr(event_handler, "on_service_interrupted"):
                self.service.add_service_interruption_listener(event_handler)

    def listen(self, handler: type[MessageHandler], queue_name: str | None = None) -> None:
        """Start listening for messages on the specified queue."""
        # NOTE: This assumes that a persistent queue already exists on the broker with the right topic subscription
        durable_exclusive_queue = Queue.durable_exclusive_queue(queue_name or self.config.queue_name)

        # Build a receiver and bind it to the durable exclusive queue
        self.receiver: PersistentMessageReceiver = (
            self.service.create_persistent_message_receiver_builder()
            .with_missing_resources_creation_strategy(MissingResourcesCreationStrategy.CREATE_ON_START)
            .build(durable_exclusive_queue)
        )
        self.receiver.start()

        # Callback for received messages
        self.receiver.receive_async(handler(self.receiver))
        log.debug("PERSISTENT receiver started... Bound to Queue [%s]", durable_exclusive_queue.get_name())

    def close(self) -> None:
        """Close the JMS service."""
        if self.receiver is None:
            msg = "Receiver was never started, nothing to close"
            raise ValueError(msg)
        if self.receiver and self.receiver.is_running():
            log.debug("Terminating receiver")
            self.receiver.terminate(grace_period=0)
        log.debug("Disconnecting Messaging Service")
        self.service.disconnect()



"""NOTAM data structures."""

import re
from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime
from itertools import batched
from typing import Any, NamedTuple, Self

import geojson  # type: ignore
from geojson import Point, Polygon

geojson.geometry.DEFAULT_PRECISION = 13


def format_dt(data: str | dict) -> datetime:
    """Format a date string or dict into a datetime object."""
    if isinstance(data, dict):
        data = data["#text"]
    return datetime.fromisoformat(data)  # type: ignore


def optional_dt(data: str | dict) -> datetime | None:
    """Like format_dt, but returns None if the key is not found."""
    try:
        return format_dt(data)
    except KeyError:
        return None


def get_raw_text(data: dict | list[dict]) -> str:
    """Extract the raw text from the NOTAM data."""
    # If both simple and formatted are given, prefer formatted since it has more data
    if isinstance(data, list):
        data = next(item for item in data if "formattedText" in item["NOTAMTranslation"])
    raw: str
    try:
        raw = data["NOTAMTranslation"]["formattedText"]["div"]
        # Replace newlines and remove all other HTML tags
        raw = raw.replace("<BR>", "\n")
        raw = re.sub(r"<.*?>", "", raw)
    except KeyError:
        raw = data["NOTAMTranslation"]["simpleText"]
    return raw.strip()


@dataclass(frozen=True)
class TextNotam:
    """Represents a textual NOTAM."""

    id: str
    number: str
    year: str
    issued: str
    location: str
    start: str
    end: str
    text: str
    raw: str
    series: str | None
    type: str | None
    affected_fir: str | None
    selection_code: str | None
    scope: str | None
    purpose: str | None
    traffic: str | None
    schedule: str | None
    upper_limit: str | None
    lower_limit: str | None
    min_fl: str | None
    max_fl: str | None
    coordinates: str | None
    radius: str | None

    @classmethod
    def from_fil(cls, data: dict[str, str]) -> Self:
        """Create a TextNotam instance from FIL data."""
        # try:
        return cls(
            id=data["@id"],
            series=data.get("series"),
            number=data["number"],
            year=data["year"],
            type=data.get("type"),
            issued=data["issued"],
            affected_fir=data.get("affectedFIR"),
            selection_code=data.get("selectionCode"),
            scope=data.get("scope"),
            purpose=data.get("purpose"),
            traffic=data.get("traffic"),
            schedule=data.get("schedule"),
            upper_limit=data.get("upperLimit"),
            lower_limit=data.get("lowerLimit"),
            min_fl=data.get("minimumFL"),
            max_fl=data.get("maximumFL"),
            coordinates=data.get("coordinates"),
            radius=data.get("radius"),
            location=data["location"],
            start=data["effectiveStart"],
            end=data["effectiveEnd"],
            text=data["text"],
            raw=get_raw_text(data["translation"]),  # type: ignore
        )
        # except KeyError as exc:
        #     # pprint(data)
        #     raise ValueError from exc


class _Features(NamedTuple):
    notes: list[str]
    shapes: list[Point | Polygon]


def _extract_features(features: _Features, data: dict[str, Any]) -> None:
    """Recursively extract specific field types from the data."""
    for key, val in data.items():
        if isinstance(val, list) and val and isinstance(val[0], dict):
            for item in val:
                _extract_features(features, item)
        elif isinstance(val, dict):
            _extract_features(features, val)
        elif not isinstance(val, str):
            continue
        elif key in ("note", "operationalStatus", "status"):
            features.notes.append(val)
        elif key == "pos":
            lat, lon = map(float, val.split(" "))
            features.shapes.append(Point((lon, lat)))
        elif key == "posList":
            coords = [(lon, lat) for lat, lon in batched(map(float, val.split(" ")), 2)]
            features.shapes.append(Polygon(coords))


def check_event(item: dict) -> dict | None:
    """There are two different event types. We want the one with the extension"""
    with suppress(KeyError):
        event: dict = item["Event"]["timeSlice"]["EventTimeSlice"]
        _ = event["extension"]
        return event
    return None


@dataclass(frozen=True)
class Notam:
    """Represents a FAA SWIFT NOTAM."""

    id: str
    issued: datetime
    updated: datetime
    start: datetime
    end: datetime | None

    classification: str
    icao: str | None
    name: str | None

    notes: list[str]
    shapes: list[Point | Polygon]
    text: TextNotam

    @classmethod
    def from_fil(cls, data: dict[str, Any]) -> Self:
        """Create a Notam instance from FIL data."""
        # try:
        root = data["hasMember"]
        event: dict[str, Any]
        features = _Features(notes=[], shapes=[])
        if isinstance(root, list):
            for item in root:
                if "Event" in item:
                    if checked_event := check_event(item):
                        event = checked_event
                        break
                else:
                    _extract_features(features, item)
        else:
            event = root["Event"]["timeSlice"]["EventTimeSlice"]
        times: dict = event["validTime"]["TimePeriod"]
        notam: dict = event["textNOTAM"]["NOTAM"]
        extension: dict = event["extension"]["EventExtension"]
        return cls(
            id=event["@id"],
            issued=format_dt(notam["issued"]),
            updated=format_dt(extension["lastUpdated"]),
            start=format_dt(times["beginPosition"]),
            end=optional_dt(times["endPosition"]),
            classification=extension["classification"],
            icao=extension.get("icaoLocation"),
            name=extension.get("airportname"),
            notes=features.notes,
            shapes=features.shapes,
            text=TextNotam.from_fil(notam),
        )
        # except Exception as exc:
        #     # pprint(data)
        #     raise ValueError from exc


# Worth a second look:

# "RadioFrequencyArea": {
#     "@id": "RFA01_66789677",
#     "boundedBy": {
#       "@nil": "true",
#       "@xmlns": {
#         "xsi": "http://www.w3.org/2001/XMLSchema-instance"
#       }
#     },
#     "timeSlice": {
#       "RadioFrequencyAreaTimeSlice": {
#         "@id": "RFA01_TS01_66789677",
#         "extension": {
#           "RadioFrequencyAreaExtension": {
#             "@id": "RFAE_EVENT_1_66789677",
#             "theEvent": {
#               "@href": "#Event_1_66789677"
#             }
#           }
#         },
#         "interpretation": "TEMPDELTA",
#         "sector": {
#           "CircleSector": {
#             "@id": "CS01_66789677",
#             "angleType": "RDL",
#             "arcDirection": "CWA",
#             "fromAngle": "228",
#             "outerDistance": {
#               "#text": "26",
#               "@uom": "NM"
#             },
#             "toAngle": "284",
#             "upperLimit": {
#               "#text": "15900",
#               "@uom": "FT"
#             },
#             "upperLimitReference": "SFC"
#           }
#         },
#         "validTime": {
#           "TimePeriod": {
#             "@id": "RFA01_TS02_TP01_66789677",
#             "beginPosition": "2023-01-05T18:51:00.000Z",
#             "endPosition": {
#               "@indeterminatePosition": "unknown"
#             }
#           }
#         }
#       }
#     }
#   },



""" "Service for interacting with the FIL SFTP server."""

import json
from datetime import UTC, datetime
from gzip import GzipFile
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Self

import xmltodict
from sftputil import SFTP

from avwx_swift.notam import Notam

DATE_FILE = "primary_last_date.txt"
# TIME_FORMAT = "%Y-%m-%d %H:%M:%S %Z"
FIL_FILE_NAME = "initial_load_aixm.xml.gz"
JSON_FILE_NAME = "filservice.json"


def _load_dt(value: str | None) -> datetime | None:
    if value is None:
        return None
    return datetime.fromisoformat(value).replace(tzinfo=UTC)


_NAMESPACES = (
    "http://www.w3.org/2001/XMLSchema-instance",
    "http://www.w3.org/1999/xhtml",
    "http://www.opengis.net/ows/1.1",
    "http://www.isotc211.org/2005/gts",
    "http://www.aixm.aero/schema/5.1/event",
    "urn:us.gov.dot.faa.aim.fns",
    "http://www.aixm.aero/schema/5.1/message",
    "http://www.opengis.net/wfs-util/2.0",
    "http://www.w3.org/1999/xlink",
    "http://www.opengis.net/wfs/2.0",
    "http://www.opengis.net/fes/2.0",
    "http://www.opengis.net/gml/3.2",
    "http://www.aixm.aero/schema/5.1/extensions/FAA/FNSE",
    "http://www.isotc211.org/2005/gco",
    "http://www.isotc211.org/2005/gmd",
    "http://www.aixm.aero/schema/5.1",
)
_NS_MAP = {ns: None for ns in _NAMESPACES}


class FilService:
    """Service for interacting with the FIL SFTP server."""

    url: str
    user: str
    cert_path: Path
    cache_dir: Path | None = None

    data: list[Notam]

    checked: datetime | None = None
    updated: datetime | None = None
    server_time: datetime | None = None

    conn: SFTP | None = None

    def __init__(self, url: str, user: str, cert_path: Path, cache_dir: Path | None = None) -> None:
        if cache_dir and not cache_dir.exists():
            cache_dir.mkdir(parents=True)
        self.url = url
        self.user = user
        self.cert_path = cert_path
        self.cache_dir = cache_dir
        self.data = []

    def __repr__(self) -> str:
        return f"FilService(size={len(self.data)}, checked={self.checked}, updated={self.updated}, server_time={self.server_time})"

    def reload_connection(self) -> None:
        """Reload the SFTP connection if any connection config changes."""
        self.conn = SFTP(self.url, self.user, key_file=self.cert_path.as_posix())

    def save_cache(self) -> None:
        """Save the current state to the cache directory."""
        if self.cache_dir is None:
            msg = "Cache directory is not set."
            raise ValueError(msg)
        attrs = {
            "url": self.url,
            "user": self.user,
            "cert_path": self.cert_path.as_posix(),
            "cache_dir": self.cache_dir.as_posix(),
            "checked": self.checked.isoformat() if self.checked else None,
            # "updated": self.updated.isoformat() if self.updated else None,
            "server_time": self.server_time.isoformat() if self.server_time else None,
        }
        cache_file = self.cache_dir / JSON_FILE_NAME
        with cache_file.open("w") as fout:
            json.dump(attrs, fout)

    def update(self, *, force: bool = False) -> bool:
        """Update the local cache with the latest data from the SFTP server."""
        if self.cache_dir is None:
            with TemporaryDirectory() as tmpdir:
                return self._update(Path(tmpdir), force=force)
        else:
            return self._update(self.cache_dir, force=force)

    def update_server_time(self) -> datetime:
        """Fetch and return the most recent update time from the SFTP server."""
        if self.cache_dir is None:
            with TemporaryDirectory() as tmpdir:
                return self._update_server_time(Path(tmpdir))
        else:
            return self._update_server_time(self.cache_dir)

    @property
    def should_update(self) -> bool:
        """Determine if an update is needed based on server time."""
        if not self.server_time:
            msg = "Server time has not been fetched."
            raise ValueError(msg)
        if not self.updated:
            return True
        return self.updated < self.server_time

    def _update(self, target_dir: Path, *, force: bool) -> bool:
        """Update the local cache with the latest data from the SFTP server."""
        self._update_server_time(target_dir)
        if not (force or self.should_update):
            return False
        self._download(target_dir)
        self._parse(target_dir)
        if self.cache_dir:
            self.save_cache()
        return True

    def _get_file(self, name: str, target: Path) -> None:
        """Download a file from the SFTP server using the command line."""
        if self.conn is None:
            self.reload_connection()
        # When caching, always overwrite existing files because SFTP won't overwrite
        local_target = target / name
        if local_target.exists():
            local_target.unlink()
        conn: SFTP = self.conn
        conn.sync_pull(name, target.as_posix())

    def _update_server_time(self, target_dir: Path) -> datetime:
        """Fetch and return the most recent update time from the SFTP server."""
        self._get_file(DATE_FILE, target_dir)
        with (target_dir / DATE_FILE).open() as fin:
            self.server_time = datetime.fromisoformat(fin.read().strip()).replace(tzinfo=UTC)
        self.checked = datetime.now(UTC)
        return self.server_time

    def _download(self, target_dir: Path) -> None:
        """Download the FIL file from the SFTP server."""
        self._get_file(FIL_FILE_NAME, target_dir)

    def _parse(self, target_dir: Path) -> None:
        """Parse the FIL file and extract NOTAMs."""
        self.data = []

        def parse_notam(_: Any, item: dict[str, Any]) -> bool:
            self.data.append(Notam.from_fil(item))
            return True  # Indicates that the parser should continue

        xmltodict.parse(
            GzipFile(target_dir / FIL_FILE_NAME),
            item_depth=5,
            item_callback=parse_notam,
            process_namespaces=True,
            namespaces=_NS_MAP,
        )

        self.updated = datetime.now(UTC)

    @classmethod
    def from_cache(cls, cache_dir: Path, *, load_data: bool = True) -> Self:
        """Load the FIL service instance from the cache directory."""
        if not cache_dir.exists():
            msg = "Cache directory does not exist."
            raise ValueError(msg)
        cache_file = cache_dir / JSON_FILE_NAME
        with cache_file.open() as fin:
            attrs = json.load(fin)
        obj = cls(
            url=attrs["url"],
            user=attrs["user"],
            cert_path=Path(attrs["cert_path"]),
            cache_dir=Path(attrs["cache_dir"]),
        )
        obj.checked = _load_dt(attrs["checked"])
        obj.server_time = _load_dt(attrs["server_time"])
        if load_data:
            obj._parse(cache_dir)  # noqa
        return obj



"""Module defining data structures for FAA FNS messages and NOTAMs."""

from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum, auto


class NotamStatus(StrEnum):
    """Enumeration for NOTAM status."""

    ACTIVE = auto()
    CANCELLED = auto()
    EXPIRED = auto()


@dataclass
class FnsMessage:
    id: int
    correlation_id: int
    issued: datetime
    updated: datetime
    valid_from: datetime
    valid_to: datetime
    classification: str
    location_designator: str

    @classmethod
    def from_dict(cls, correlation_id: int, data: dict) -> "FnsMessage":
        """Create an instance from a dictionary."""
        raise NotImplementedError


class FnsNotam(FnsMessage):
    """A NOTAM message from the FAA FNS system."""

    text: str
    status: NotamStatus
    accountability: str
    aixm_message: str

    @classmethod
    def from_dict(cls, correlation_id: int, data: dict) -> "FnsNotam":
        """Create an instance from a dictionary."""
        raise NotImplementedError
